module my_data
  implicit none
  INTEGER, ALLOCATABLE:: IAP(:), JA(:)
  Double precision, ALLOCATABLE:: A(:), DA(:,:)
  Integer mode
  integer, allocatable:: start_row(:)
  Integer m, opcount
  Double precision s0,s1
  INTEGER LWORK
  DOUBLE PRECISION, ALLOCATABLE:: WORK(:)
end module my_data

SUBROUTINE DTRF(K,L,P,VK,M,CPTM,BE)
  IMPLICIT NONE

  INTEGER K,L,I,J,M
  DOUBLE PRECISION TWO
  PARAMETER ( TWO=2.0D+0 )
  DOUBLE PRECISION P(K,L),CPTM(K,K),VK(M,K),BE(1,K)
  DOUBLE PRECISION CS,SN,PP,PQ,QQ
  
  DO I=1,L
     DO J=K,I+1,-1
        CALL DLARTG(P(J-1,I),P(J,I),CS,SN,P(J-1,I))

        CALL DROT(L-I,P(J-1,I+1),K,P(J,I+1),K,CS,SN)

        PP=CPTM(J-1,J-1)
        PQ=CPTM(J-1,J)
        QQ=CPTM(J,J)
        
        CALL DROT(K,CPTM(1,J-1),1,CPTM(1,J),1,CS,SN)
        CALL DROT(K,CPTM(J-1,1),K,CPTM(J,1),K,CS,SN)
        
        CPTM(J-1,J) = CS*SN*(QQ-PP) &
             +PQ*(CS-SN)*(CS+SN)
        CPTM(J,J-1) = CPTM(J-1,J)
        CPTM(J-1,J-1) = TWO*CS*SN*PQ &
             +CS**2*PP+SN**2*QQ
        CPTM(J,J) = -TWO*CS*SN*PQ &
             +CS**2*QQ+SN**2*PP

        CALL DROT(M,VK(1,J-1),1,VK(1,J),1,CS,SN)
        CALL DROT(1,BE(1,J-1),1,BE(1,J),1,CS,SN)
     ENDDO
  ENDDO
  
  RETURN
END SUBROUTINE DTRF

SUBROUTINE RESL(J,LS,L,K,TM,VK,VPLUS,METHOD)
  use my_data
  use omp_lib
  IMPLICIT NONE

  INTEGER LS

  INTEGER K,L,I,J,JJ,IINFO,METHOD
  DOUBLE PRECISION ONE,ZERO,TWO
  PARAMETER(ONE=1.0D+0,ZERO=0.0D+0,TWO=2.0D+0)
  DOUBLE PRECISION ALPHA,BETA
  DOUBLE PRECISION TM(K,K),CPTM(K,K),CPTM2(K,K),VK(M,K),VPLUS(M),BD(K),BE(K)
  DOUBLE PRECISION VM(M,L)
  DOUBLE PRECISION WORK2(K*8)
  DOUBLE PRECISION P(K,K)
  DOUBLE PRECISION DDOT,DNRM2

  VK(1:M,J+1) = VPLUS(1:M)
  DO WHILE(J < K)
     CALL av(VK(1:M,J+1),VPLUS)

     ALPHA = DDOT(M,VPLUS,1,VK(1:M,J+1),1)
     TM(J+1,J+1) = ALPHA
     CALL DAXPY(M,-ALPHA,VK(1:M,J+1),1,VPLUS,1)

     ALPHA = DDOT(M,VPLUS,1,VK(1:M,J+1),1)
     TM(J+1,J+1) = TM(J+1,J+1)+ALPHA
     CALL DAXPY(M,-ALPHA,VK(1:M,J+1),1,VPLUS,1)

     CALL CGS2(VPLUS,VK,M,J,WORK)

     BETA = DNRM2(M,VPLUS,1)
     IF(J < K-1) THEN
        TM(J+1,J+2) = BETA
        VK(1:M,J+2) = VPLUS(1:M) / BETA
     ELSE
        VPLUS(1:M) = VPLUS(1:M) / BETA
     END IF
     J = J+1
  END DO
  J = L

  P = ZERO
  DO I = 1,K
     P(I,I) = ONE
  END DO

  IF (METHOD == 1) THEN
     
     DO I= 2, K
        DO JJ=1,I-1
           TM(I,JJ)=TM(JJ,I)
        END DO
     END DO
     
     CPTM = TM
     CALL DGEBRDG_4_TRISIDE(L+1,CPTM,K,P)
     
     DO I = 1,K
        BD(I) = CPTM(I,I)
     END DO
     DO I = 1,K-1
        BE(I) = CPTM(I,I+1)
     END DO
     CALL DSTEQR('V', K, BD, BE, P, K, WORK, IINFO ) 
     CALL userorder(LS,K,BD,P)
     
     CALL DGEMM('N','N',M,L,K,ONE,VK,M,P,K,ZERO,VM,M)
     CALL DLACPY('A',M,L,VM,M,VK,M)
     
     TM = ZERO
     DO I = 1,L
        TM(I,I) = BD(I)
     END DO
     CALL DAXPY(L,BETA,P(K,1),K,TM(1,L+1),1)
     
  ELSE IF (METHOD == 2) THEN
     
     DO I= 2, K
        DO JJ=1,I-1
           TM(I,JJ)=TM(JJ,I)
        END DO
     END DO
     
     CPTM = TM
     CALL DGEBRDG_4_TRISIDE(L+1,TM,K,P)
     
     DO I = 1,K
        BD(I) = TM(I,I)
     END DO
     DO I = 1,K-1
        BE(I) = TM(I,I+1)
     END DO
     CALL DSTEQR('V', K, BD, BE, P, K, WORK, IINFO ) 
     CALL userorder(LS,K,BD,P)
     
     CALL DGEQR2P(K,L,P,K,BD,WORK,IINFO)
     CALL DORM2R('R','N',M,K,L,P,K,BD,VK,M,WORK,IINFO )

     BE(1:K) = ZERO
     BE(K) = BETA
     CALL DORM2R('R','N',1,K,L,P,K,BD,BE,1,WORK,IINFO ) 
     
     CALL DORM2R('L','T',K,K,L,P,K,BD,CPTM,K,WORK,IINFO )
     CALL DORM2R('R','N',L,K,L,P,K,BD,CPTM,K,WORK,IINFO )
     
     TM = ZERO
     DO I= 1, L
        TM(I,I)=CPTM(I,I)
     END DO
     
     CALL DCOPY(L,BE,1,TM(1,L+1),1)
     
  ELSE IF (METHOD == 3) THEN
     
     DO I= 2, K
        DO JJ=1,I-1
           TM(I,JJ)=TM(JJ,I)
        END DO
     END DO
     
     CPTM = TM
     CALL DGEBRDG_4_TRISIDE(L+1,TM,K,P)
     
     DO I = 1,K
        BD(I) = TM(I,I)
     END DO
     DO I = 1,K-1
        BE(I) = TM(I,I+1)
     END DO
     CALL DSTEQR('V', K, BD, BE, P, K, WORK, IINFO ) 
     CALL userorder(LS,K,BD,P)
     
     BE(1:K) = ZERO
     BE(K) = BETA

     CALL DTRF(K,L,P,VK,M,CPTM,BE)
     
     TM = ZERO
     DO I= 1, L
        TM(I,I)=CPTM(I,I)
     END DO
     
     CALL DCOPY(L,BE,1,TM(1,L+1),1)
     
  ELSE IF (METHOD == 4) THEN
     
     DO I= 2, K
        DO JJ=1,I-1
           TM(I,JJ)=TM(JJ,I)
        END DO
     END DO

     CPTM = TM
     CALL DGEBRDG_4_TRISIDE(L+1,TM,K,P)

     DO I = 1,K
        BD(I) = TM(I,I)
     END DO
     DO I = 1,K-1
        BE(I) = TM(I,I+1)
     END DO
     CALL DSTEQR('V', K, BD, BE, P, K, WORK, IINFO )
     CALL userorder(LS,K,BD,P)

     CALL MGS(K,L,P,WORK2)
     CALL DGEMM('N','N',M,L,K,ONE,VK,M,P,K,ZERO,VM,M)
     CALL DLACPY('A',M,L,VM,M,VK,M)

     CALL DSYMM('L','U',K,L,ONE,CPTM,K,P,K,ZERO,CPTM2,K)
     TM = ZERO
     DO I = 1,L
        TM(I,I) = DDOT(K,CPTM2(1,I),1,P(1,I),1)
     END DO

     CALL DAXPY(L,BETA,P(K,1),K,TM(1,L+1),1)

  END IF
  
  RETURN
END SUBROUTINE RESL

subroutine userorder(LS,K,BD,TM)

  IMPLICIT NONE
  include 'omp_lib.h'
  INTEGER II,JJ,Z,LS,K,I
  DOUBLE PRECISION TM(K,K),BD(K),PP,PQ

  if (LS == 0) then
     DO II = 2, K
        I = II - 1
        Z = I
        PP= abs(BD( I ))
        DO JJ = II, K
           IF(abs( BD( JJ )).GT.PP) THEN
              Z = JJ
              PP= abs(BD( JJ ))
           END IF
        END DO
        IF( Z.NE.I ) THEN
           PQ=BD(Z)
           BD( Z ) = BD( I )
           BD( I ) = PQ
           CALL DSWAP( K, TM( 1, I ), 1, TM( 1, Z ), 1 )
        END IF
     END DO
  else if (LS == 1) then
     DO II = 2, K
        I = II - 1
        Z = I
        PP= abs(BD( I ))
        DO JJ = II, K
           IF(abs( BD( JJ )).LT.PP) THEN
              Z = JJ
              PP= abs(BD( JJ ))
           END IF
        END DO
        IF( Z.NE.I ) THEN
           PQ=BD(Z)
           BD( Z ) = BD( I )
           BD( I ) = PQ
           CALL DSWAP( K, TM( 1, I ), 1, TM( 1, Z ), 1 )
        END IF
     END DO
  else if (LS == 2) then
     DO II = 2, K
        I = II - 1
        Z = I
        PP= BD( I )
        DO JJ = II, K
           IF( BD( JJ ).GT.PP) THEN
              Z = JJ
              PP= BD( JJ )
           END IF
        END DO
        IF( Z.NE.I ) THEN
           PQ=BD(Z)
           BD( Z ) = BD( I )
           BD( I ) = PQ
           CALL DSWAP( K, TM( 1, I ), 1, TM( 1, Z ), 1 )
        END IF
     END DO
  end if

end subroutine userorder
!
Subroutine av(xin, yout)
  Use my_data
  Use omp_lib
  Implicit None
  Integer i, j,thr_num
  Double precision, Dimension(m), Intent(in) :: xin
  Double precision, Dimension(m), Intent(out) :: yout
  Double precision, ALLOCATABLE:: yout_tmp(:)
  Double precision  zero
  parameter (zero=0.0d0)
  EXTERNAL           DDOT
  double precision   DDOT

  s1 = omp_get_wtime()
  IF (.FALSE.) THEN

     if (mode .eq. 0) then

        do i=1,m
           yout(i) = zero
        enddo

!$OMP PARALLEL PRIVATE(i,thr_num,j,yout_tmp)
        ALLOCATE (yout_tmp(m))

        do i=1,m
           yout_tmp(i) = zero
        enddo
        
        thr_num = omp_get_thread_num()
        do i=start_row(thr_num+1),start_row(thr_num+2)-1
           do j=IAP(i),IAP(i+1)-1
              yout_tmp(i) = yout_tmp(i)+(A(j))*xin(JA(j))
              yout_tmp(JA(j)) = yout_tmp(JA(j))+(A(j))*xin(i)
           enddo
        enddo

        do j=1,omp_get_max_threads()
!$OMP barrier
           if (j == thr_num) then
              do i=1,m
                 yout(i)=yout(i)+yout_tmp(i)
              enddo
           endif
        enddo

        DEALLOCATE (yout_tmp)
!$OMP END PARALLEL

     else
        
        do i=1,m
           yout(i) = zero
        enddo

!$OMP PARALLEL PRIVATE(i,thr_num,j,yout_tmp)
        ALLOCATE (yout_tmp(m))

        do i=1,m
           yout_tmp(i) = zero
        enddo

!$OMP DO
        do i=1,m
           do j=1,m
              yout_tmp(i) = yout_tmp(i)+(DA(j,i))*xin(j)
              yout_tmp(j) = yout_tmp(j)+(DA(j,i))*xin(i)
           enddo
        enddo
!$OMP END DO

        thr_num = omp_get_thread_num()
        do j=1,omp_get_max_threads()
!$OMP barrier
           if (j == thr_num) then
              do i=1,m
                 yout(i)=yout(i)+yout_tmp(i)
              enddo
           endif
        enddo

        DEALLOCATE (yout_tmp)
!$OMP END PARALLEL

     end if
     
  ELSE
  
     if (mode .eq. 0) then

!$OMP PARALLEL PRIVATE(i,thr_num,j)
        thr_num = omp_get_thread_num()
        do i=start_row(thr_num+1),start_row(thr_num+2)-1
           yout(i) = zero
           do j=IAP(i),IAP(i+1)-1
              yout(i) = yout(i)+(A(j))*xin(JA(j))
           enddo
        enddo
!$OMP END PARALLEL
     
     else

!$OMP PARALLEL DO
        do i=1,M
           yout(i) = ddot(m,DA(1,i),1,xin,1)
        enddo
!$OMP END PARALLEL DO

     end if
  endif
  s0 = s0 + (omp_get_wtime() - s1)
  
  opcount = opcount + 1
  return
End Subroutine av
